#include "stdafx.h"
#include "VyvojovaDoska.h"
#include <lusb0_usb.h>

#pragma comment(lib, "libusb.lib")

// VID zariadenia
#define DEVICE_VID 0xABCD
// PID zariadenia
#define DEVICE_PID 0x1234

// Konfigurciu, ktor zvolme
#define DEVICE_CONFIG 1
// Interface, ktor zvolme
#define DEVICE_INTERFACE 0
// Endpoint pre prenos BULK OUT
#define DEVICE_EP_BULK_OUT 0x06
// Endpoint pre prenos INTERRUPT IN
#define DEVICE_EP_INTERRUPT_IN 0x83
// Vekos bufferu pre BULK prenos
#define DEVICE_BULK_BUFFER_SIZE 16212
// Max. vekos textu
#define DEVICE_MAX_TEXT_LENGTH 20
// rka obrazovky
#define DEVICE_LCD_WIDTH 240
// Vka obrazovky
#define DEVICE_LCD_HEIGHT 320
// Max. vekos bitmapy
#define DEVICE_MAX_BITMAP_SIZE 8100
// Vekos bloku bitmapy
#define DEVICE_DISPLAY_BLOCK 90u

// Intancia otvorenho zariadenia
usb_dev_handle * deviceHandle = 0;
// Buffer pre BULK prenos
char deviceBulkBuffer[DEVICE_BULK_BUFFER_SIZE];

// Skonfiguruje otvoren zariadenie
unsigned int ConfigureDevice()
{
	if (!deviceHandle)
		return ResultOpen_Error;

	// Zvoli konfigurciu
	if (usb_set_configuration(deviceHandle, DEVICE_CONFIG) < 0)
		return ResultOpen_Configuration;
	// Zvoli rozhranie
	if (usb_claim_interface(deviceHandle, DEVICE_INTERFACE) < 0)
		return ResultOpen_Interface;
	// Zvoli alternatvne rozhranie, zariadenie sa aktivuje
	if (usb_set_altinterface(deviceHandle, 1) < 0)
		return ResultOpen_AltInterface;

	return ResultOpen_OK;
}

// Otvor intanciu zariadenia "Vvojov doska"
__declspec(dllexport) unsigned int OpenDevice()
{
	if (deviceHandle)
		return ResultOpen_Error;

	struct usb_bus * bus;
	struct usb_device * device;

	// Njs vetky zbernice a zariadenia
	usb_find_busses();
	usb_find_devices();

	// Pre kad USB zbernicu
	for (bus = usb_get_busses(); bus; bus = bus->next)
	{
		// Pre kad zariadenie na zbernici
		for (device = bus->devices; device; device = device->next)
		{
			// Porovna VID a PID zariadenia s naim
			if (device->descriptor.idVendor == DEVICE_VID &&
				device->descriptor.idProduct == DEVICE_PID)
			{
				// Otvori intanciu zariadenia
				deviceHandle = usb_open(device);

				if (deviceHandle)
				{
					// Skonfigurova zariadenie
					unsigned int config = ConfigureDevice();
					if (config != ResultOpen_OK)
						CloseDevice();

					return config;
				}
				else
					return ResultOpen_ErrorOpen;
			}
		}
	}

	return ResultOpen_NotFound;
}

// Zatvor otvoren intanciu zariadenia
__declspec(dllexport) void CloseDevice()
{
	if (deviceHandle)
	{
		// Zatvori intanciu zariadenia
		usb_close(deviceHandle);
		deviceHandle = 0;
	}
}

// Preta hodnoty z prevodnka ADC (potenciometer a teplotu)
__declspec(dllexport) unsigned int ReadADC(usbapp_control_in_t * data)
{
	if (!deviceHandle)
		return DeviceError_NotOpen;
	if (!data)
		return DeviceError_InvalidParameter;

	// Odosla CONTROL IN paket
	int ret = usb_control_msg(deviceHandle,
		USB_TYPE_VENDOR | USB_RECIP_INTERFACE | USB_ENDPOINT_IN,
		0, ControlType_In, DEVICE_INTERFACE,
		(char *)data, sizeof(usbapp_control_in_t), 0);

	if (ret < 0)
		return DeviceError_Transfer;
	if (ret != sizeof(usbapp_control_in_t))
		return DeviceError_InvalidSize;

	return DeviceError_OK;
}

// Nastav zariadenie (LED a LCD)
__declspec(dllexport) unsigned int SetupDevice(usbapp_control_out_t * data)
{
	if (!deviceHandle)
		return DeviceError_NotOpen;
	if (!data)
		return DeviceError_InvalidParameter;

	// Opravi hodnoty v parametroch
	if (data->pwmLedDuty > usbapp_control_pwm_max)
		data->pwmLedDuty = usbapp_control_pwm_max;

	// Odosla CONTROL OUT paket
	int ret = usb_control_msg(deviceHandle,
		USB_TYPE_VENDOR | USB_RECIP_INTERFACE | USB_ENDPOINT_OUT,
		0, ControlType_Out, DEVICE_INTERFACE,
		(char *)data, sizeof(usbapp_control_out_t), 0);

	if (ret < 0)
		return DeviceError_Transfer;
	if (ret != sizeof(usbapp_control_out_t))
		return DeviceError_InvalidSize;

	return DeviceError_OK;
}

// Pok na vstupn paket preruenia (ke sa niekto dotkne tlaidiel na doske)
__declspec(dllexport) unsigned int WaitForInterrupt(usbapp_interrupt_in_t * data)
{
	if (!deviceHandle)
		return DeviceError_NotOpen;
	if (!data)
		return DeviceError_InvalidParameter;

	// Prija INTERRUPT IN
	int ret = usb_interrupt_read(deviceHandle, DEVICE_EP_INTERRUPT_IN,
		(char *)data, sizeof(usbapp_interrupt_in_t), 0);

	if (ret < 0)
		return DeviceError_Transfer;
	if (ret != sizeof(usbapp_interrupt_in_t))
		return DeviceError_InvalidSize;

	return DeviceError_OK;
}

// Vyist obrazovku jednou farbou
__declspec(dllexport) unsigned int ClearScreen(unsigned int color)
{
	if (!deviceHandle)
		return DeviceError_NotOpen;

	// Nastavi parametre hlaviky
	usbapp_bulk_out_t * data = (usbapp_bulk_out_t *)deviceBulkBuffer;
	memset(data, 0, sizeof(usbapp_bulk_out_t));
	data->options = DrawOption_Clear | (color & DrawOption_ColorMask);

	// Odosla BULK OUT paket
	int ret = usb_bulk_write(deviceHandle, DEVICE_EP_BULK_OUT,
		deviceBulkBuffer, sizeof(usbapp_bulk_out_t), 0);

	if (ret < 0)
		return DeviceError_Transfer;
	if (ret != sizeof(usbapp_bulk_out_t))
		return DeviceError_InvalidSize;

	return DeviceError_OK;
}

// Vykresl text na obrazovke
__declspec(dllexport) unsigned int DrawString(usbapp_bulk_out_t * options, const char * text,
	unsigned int textLength)
{
	if (!deviceHandle)
		return DeviceError_NotOpen;
	if (!options)
		return DeviceError_InvalidParameter;
	if (!text)
		return DeviceError_InvalidParameter;
	if (textLength > DEVICE_MAX_TEXT_LENGTH || textLength == 0)
		return DeviceError_InvalidParameter;

	// Opravi hodnoty v parametroch
	usbapp_bulk_out_t * data = (usbapp_bulk_out_t *)deviceBulkBuffer;
	memcpy_s(deviceBulkBuffer, DEVICE_BULK_BUFFER_SIZE,
		options, sizeof(usbapp_bulk_out_t));
	if (data->posX >= DEVICE_LCD_WIDTH)
		data->posX = 0;
	if (data->posY >= DEVICE_LCD_HEIGHT)
		data->posY = 0;
	data->options |= DrawOption_Text;
	data->options &= ~(DrawOption_Bitmap);

	// Skoprova text za hlaviku prenosu
	memcpy_s(deviceBulkBuffer + sizeof(usbapp_bulk_out_t),
		DEVICE_BULK_BUFFER_SIZE - sizeof(usbapp_bulk_out_t),
		text, textLength);
	
	int size = sizeof(usbapp_bulk_out_t) + textLength;

	// Odosla BULK OUT paket
	int ret = usb_bulk_write(deviceHandle, DEVICE_EP_BULK_OUT,
		deviceBulkBuffer, size, 0);

	if (ret < 0)
		return DeviceError_Transfer;
	if (ret != size)
		return DeviceError_InvalidSize;

	return DeviceError_OK;
}

// Zska adresu, kde je mon vloi bitmapu na vykreslenie
__declspec(dllexport) uint16_t * GetBitmapBuffer()
{
	return (uint16_t *)(deviceBulkBuffer + sizeof(usbapp_bulk_out_t));
}

// Vykresl as obrzku na displej
__declspec(dllexport) unsigned int DrawBitmap(usbapp_bulk_out_t * options)
{
	if (!deviceHandle)
		return DeviceError_NotOpen;
	if (!options)
		return DeviceError_InvalidParameter;
	if ((options->height * options->width) > DEVICE_MAX_BITMAP_SIZE)
		return DeviceError_InvalidParameter;
	if (options->width > DEVICE_LCD_WIDTH ||
		options->height > DEVICE_LCD_HEIGHT)
		return DeviceError_InvalidParameter;

	// Opravi hodnoty v parametroch
	usbapp_bulk_out_t * data = (usbapp_bulk_out_t *)deviceBulkBuffer;
	memcpy_s(deviceBulkBuffer, DEVICE_BULK_BUFFER_SIZE,
		options, sizeof(usbapp_bulk_out_t));
	if (data->posX >= DEVICE_LCD_WIDTH)
		data->posX = 0;
	if (data->posY >= DEVICE_LCD_HEIGHT)
		data->posY = 0;
	data->options &= ~(DrawOption_Text);
	
	int size = sizeof(usbapp_bulk_out_t) + 
		((options->height * options->width) * 2);

	// Odosla BULK OUT paket
	int ret = usb_bulk_write(deviceHandle, DEVICE_EP_BULK_OUT,
		deviceBulkBuffer, size, 0);

	if (ret < 0)
		return DeviceError_Transfer;
	if (ret != size)
		return DeviceError_InvalidSize;

	return DeviceError_OK;
}

// Vykresl cel obrzok na displej
__declspec(dllexport) unsigned int DrawImage(unsigned int x, unsigned int y,
	unsigned int width, unsigned int height, uint16_t * data)
{
	if (!deviceHandle)
		return DeviceError_NotOpen;
	if (x >= usbapp_display_width || y >= usbapp_display_height ||
		width == 0 || height == 0 || !data)
		return DeviceError_InvalidParameter;

	// Inicializova hlaviku BULK prenosu
	usbapp_bulk_out_t options;
	memset(&options, 0, sizeof(usbapp_bulk_out_t));
	options.posX = x;
	options.posY = y;
	options.options = DrawOption_Bitmap;

	// Nesmieme prena obrzky, ktor by presiahli rozmery LCD displeja
	int _width = (x + width) > usbapp_display_width ?
		(usbapp_display_width - x) : width;
	int _height = (y + height) > usbapp_display_height ?
		(usbapp_display_height - y) : height;

	// Celkov vka v px, ktor ete potrebujeme spracova
	unsigned int totalHeight = _height;
	// Celkov rka v px, ktor ete potrebujeme spracova
	unsigned int totalWidth;
	// Prenesen rka; prenesen vka; kd poslednej chyby
	unsigned currentWidth, currentHeight, error;
	// Koko px rky a vky sme preniesli
	unsigned int usedWidth = 0, usedHeight = 0;

	// Adresa pamte, do ktorej ukladme bitmapu na prenos
	uint16_t * buffer = GetBitmapBuffer();

	// Pokia je nespracovan vka v px via ako 0
	while (totalHeight > 0)
	{
		// Nastavi premenn pre al "riadok"
		options.posX = x;
		totalWidth = _width;
		usedWidth = 0;

		// Vypota vku bloku, ktor budeme prena
		currentHeight = totalHeight > DEVICE_DISPLAY_BLOCK ?
			DEVICE_DISPLAY_BLOCK : totalHeight;

		// Pokia je nespracovan rka v px via ako 0
		while (totalWidth > 0)
		{
			// Vypota rku bloku, ktor budeme prena
			currentWidth = totalWidth > DEVICE_DISPLAY_BLOCK ?
				DEVICE_DISPLAY_BLOCK : totalWidth;

			// Nastavi v hlavike rku a vku aktulne prenanho bloku
			options.width = currentWidth;
			options.height = currentHeight;

			// Skoprova as obrzku do zsobnka dt
			for (unsigned int i = 0; i < currentHeight; i++)
				memcpy(buffer + (i * currentWidth), data + (((i + usedHeight) * width) + usedWidth), currentWidth * 2);

			// Odosla jeden blok
			error = DrawBitmap(&options);
			if (error != DeviceError_OK)
				return error;

			// Aktualizova premenn
			totalWidth -= currentWidth;
			options.posX += currentWidth;
			usedWidth += currentWidth;
		}

		// Aktualizova premenn
		totalHeight -= currentHeight;
		options.posY += currentHeight;
		usedHeight += currentHeight;
	}

	return DeviceError_OK;
}

#define rgb24_to_rgb16(ul_color) (((ul_color >> 8) & 0xF800) | ((ul_color >> 5) & 0x7E0) | ((ul_color >> 3) & 0x1F))

// Skonvertuje 32 bitov bitmapu na 16 bitov
__declspec(dllexport) void Convert32Bitmap(uint32_t * source, uint16_t * destination,
	unsigned int width, unsigned int height)
{
	if (source && destination && width > 0 && height > 0)
	{
		for (unsigned int i = 0; i < height; i++)
		{
			for (unsigned int e = 0; e < width; e++)
				destination[i * width + e] = rgb24_to_rgb16(source[i * width + e]);
		}
	}
}